@@ -1,9 +1,9 @@
 # Name:: FormatR
 # Description:: Perl like formats for ruby
-# Author:: Paul Rubel (rubel@crhc.uiuc.edu)
-# Release:: 1.06
-# Homepage:: http://www.crhc.uiuc.edu/~rubel/FormatR.html
-# Date:: 29 July 2002
+# Author:: Paul Rubel (prubel@sourceforge.net)
+# Release:: 1.07
+# Homepage:: http://formatr.sourceforge.net
+# Date:: Dec 30 2002
 # License:: You can redistribute it and/or modify it under the same term as Ruby.
 #           Copyright (c) 2002  Paul Rubel
 #
@@ -91,13 +91,13 @@
 #   of your specification (regarding the whole as more important than the
 #   fraction). 
 #
-# * A line that starts with ~ will be suppressed if it will be blank
+# * A line that contains a ~ will be suppressed if it will be blank
 #
 # * A line that contains ~~ will repeat until it is blank, be sure
 #   to use this feature with at least one field starting with a ^.
 #
 # === Scientific formats of the form @.#G##, @.#g##, @.#E##, and @.#e##
-# *  The use of G, g, E, and e is consistent to their use in printf. 
+# *  The use of G, g, E, and e is consistent with their use in printf. 
 #
 # * If a G or g is specified the number of characters before the
 #   exponent, excluding the decimal point, will give the number of
@@ -182,7 +182,7 @@
 #   
 #   # make a new FormatReader
 #   reader = FormatReader.new (format)
-#   # read in the values
+#   # Read in the values
 #    res = reader.readFormat (output)
 # 
 #   # Check that the values are correct
@@ -196,6 +196,26 @@
 #
 # = Changes:
 # 
+# ==1.07
+# * You can now use formats without having to use eval. If you pass in
+#   a hash of names to values that can be used instead. There is also 
+#   an optimization you can use by calling format.useHash(true) that
+#   will turn your binding into a hash while the format is being
+#   printed. This may speed things up. The default is still to use
+#   eval so that things do not break as some dynamic formats may not
+#   work with a hash. When a value is computed using side effects of
+#   some other evaluation that has taken place while printing the
+#   format a hash won't work. You can also use the printFormatWithHash
+#   method is you want to avoid evaling entirely. test_four in
+#   test_format.rb shows one example of how to use hashes to print formats
+#
+# * Page numbers are now working correctly. Before if you had a page
+#   number in a header or footer it was problematic. The printing of 
+#   a page has been refactored and now works much better. 
+#
+# * Thanks to Amos Gouaux for suggesting the setLinesLeft method!
+#     
+#
 # ==1.06
 # * I thought that the ~ had to be in the front of the picture line,
 #   this isn't so. If you place the ~~ anywhere in the line it will
@@ -310,9 +330,6 @@
 #
 # = Incompatibilities/Issues
 #
-# * The first format printed on a page won't see any changes done on 
-#   account of printing the top. This needs to be fixed.
-#
 # * If you use bottom be sure to check that you're happy with the
 #   output. It doesn't currently work with variable sized bottoms. You
 #   can use the finishPageWith{out}FF(...) methods to print out a
@@ -336,6 +353,10 @@
 #
 #
 # =To Do/Think about:
+# * Mark so that a user can set whether to use or not FF
+#
+# * Watch out for vars that aren't assigned but try to be used.
+#
 # * blank out undefined @##.# values with ~
 # 
 # *  use rbconfig to install?
@@ -354,15 +375,21 @@
 #    trouble if it depends upon the body format. I'm currently planning
 #    to just live with fixed sized bottoms.
 #
+# * The solution to this is probably to buffer the changes to the binding
+#    until you know they will work. I think we should pass in a hask anyways
+#    and perhaps provide a method to set variables to values in the hash. 
+#
 # * Page numbering is messy and doesn't play well with buffering lines
 #   that go on the next page. Perl gets away from this by writing more
 #   lines than will fit on a page when a format overflows. Because we cache
 #   them for later their values may look "old" if they go after a top that 
 #   was evaluated later but printed earlier. This needs help!
+#   The above point may help with this too!
 #
 # =Thanks go to
 # Hugh Sasse for his enlightening comments and suggestions. He has been incredibly
-# helpful in making this package usable.
+# helpful in making this package usable. Amos Gouaux has also been
+# helpful with suggestions and code. Thanks to both of you.
 
 module FormatR
 
@@ -378,7 +405,7 @@ module FormatR
     def initialize (val, unchanging)
       @unchanging = unchanging
       @val = val
-      if (!unchanging)
+      unless (unchanging)
         s = val.size - 1
         
         if (val =~ /[@^][<]{#{s},#{s}}/)
@@ -440,22 +467,37 @@ module FormatR
       if (max_space != nil)
         result = var_value[0,max_space]
       end
-      s = "#{var_name} = #{var_name}[#{result.size()},#{var_value.size()}].gsub(/^\\s*/,'');" +
-        "if #{var_name} == nil then #{var_name} = ''; end"
-      eval(s, aBinding)
+      setVarValue( var_name, var_value[result.size(),var_value.size()],
+                  aBinding)
       return result
     end
     
+    # Move the call to eval into one place
+    # we shouldn't have to 
+    def setVarValue (var_name, var_value, aBinding)
+      if var_value.nil? 
+        var_value = ''
+      end
+      var_value.gsub!(/^\s*/,'')
+      if (aBinding.class != binding.class) 
+        aBinding[var_name] = var_value
+      else
+        to_eval = "#{var_name} = '#{var_value}'";
+        #puts "going to eval '#{to_eval}'"
+        eval(to_eval, aBinding)
+      end
+    end
+
     # return a formatted string of the correct length
-    def formatString (var_value, var_name, binding)
+    def formatString (var_value, var_name, aBinding)
       result = var_value[0,@len]
       if (! @filled)
         return result
       end
-      return changeVarValue(var_value, var_name, binding)
+      return changeVarValue(var_value, var_name, aBinding)
     end
 
-  end
+  end #enf of class Formatter
 
   # this format doesn't care if it's a @ or an ^, it acts the same and doesn't chop things
   # used for @##.## formats
@@ -470,12 +512,12 @@ module FormatR
     # given a string that's a number spit it back with the right number of digits
     # and rounded the correct amount.
     def formatString (s, unused_var_name=nil, unused_aBinding=nil)
-      num = s.split('.') # should this take into account internationalization?
       if (s.size == 1)
         return formatInt(s)
       end
-
+      num = s.split('.') # should this take into account internationalization?
       res = num[0]
+      res = "" if (res.nil?) ## pgr xxx
       spaceLeft = @fraction + @radix
       if (res.size > @whole)
         spaceLeft = @len - res.size()
@@ -511,7 +553,9 @@ module FormatR
     end
   end
 
+  ############################################################
   # make a formatter that will spit out scientific notation
+  ############################################################
   class ScientificNotationFormatter < Formatter
     # Make a new formatter that will print out in scientific notation
     def initialize (whole, radix, fraction, precision_g_e, exponent)
@@ -535,6 +579,7 @@ module FormatR
       super
     end
 
+    #send things left
     def formatString (s, var_name, binding)
       s = super 
       s.ljust(@len)
@@ -548,6 +593,7 @@ module FormatR
       @len = val.size()
     end
 
+    #send things right
     def formatString (s, var_name, binding)
       s = super 
       s.rjust(@len)
@@ -561,6 +607,7 @@ module FormatR
       @len = val.size()
     end
 
+    #center things
     def formatString (s, var_name, binding)
       s = super
       s.center(@len)
@@ -583,8 +630,31 @@ module FormatR
     # is going even if their formats have a different io when they're
     # not attached.
     def printFormat (aBinding, io = @io)
+      if (@use_hash)
+        if (aBinding.is_a?( Binding ))
+          printFormatFromBinding( aBinding, io )                  
+        else
+          printFormatWithHash( aBinding, io )
+        end
+      else
+        printFormatWithBinding( aBinding, io )
+
+      end
+    end
+
+    # print the format given that the binding is a hash of
+    # values. This method will not call eval at all. 
+    def printFormatWithHash (aHash, io = @io)
+      useHash( true )
+      @binding = aHash
+      printBodyFormat(io)
+    end
+
+    #print the format given that the binding is actually a binding. 
+    def printFormatWithBinding (aBinding, io = @io)
+      useHash( false )
       @binding = aBinding
-      return printBodyFormat (io)
+      printBodyFormat(io)
     end
 
     # When you don't want anymore on this page just fill it with blank
@@ -604,8 +674,14 @@ module FormatR
 
     # Return how many times the top has been printed. You can use this
     # to number pages. An empty top can be used if you need the page
-    # number but don't want to print a header
+    # number but don't want to print any other header. This is a somewhat 
+    # interesting function as the bottom is only printed when a page is
+    # finished or a top is needed. If this is the case we'll pretend the
+    # page number is one h
     def pageNumber ()
+      if (@print_bottom)
+        return @page_number + 1
+      end
       return @page_number
     end
 
@@ -619,14 +695,15 @@ module FormatR
     # format you can specify it here. It will be printed once above
     # the format it is being set within. You can pass in either a
     # format or the specification of a format and it will make one for you.
-
     def setTop (format)
       top_format = format
-      if (format.class != Format)
+      if (!format.is_a?(Format))
         top_format = Format.new(format)
       end
       raise FormatException.new(), "recursive format not allowed" if (top_format == self)
       @top = top_format 
+      #in case we've already set use_hash
+      useHash( @use_hash )
 
     end
 
@@ -638,11 +715,14 @@ module FormatR
     # finishPageWith[out]FF methods.
     def setBottom (format)
       bottom_format = format
-      if (format.class != Format)
+      if (!format.is_a?(Format))
         bottom_format = Format.new(format);
       end
       raise FormatException, "recursive format not allowed" if (bottom_format == self)
       @bottom = bottom_format
+      #in case we've already set use_hash
+      useHash( @use_hash )
+
     end
 
     # Sets the number of lines on a page. If you don't want page breaks
@@ -658,15 +738,22 @@ module FormatR
     # Defaults to 60.
     def resetPage ()
       @lines_left = @page_length
+      @top.resetPage unless @top.nil?
+      @bottom.resetPage unless @bottom.nil?
     end
 
     # If you're writing to the file handle in another way than by
     # calling printFormat you can keep the pagination working using
     # this call to correctly keep track of lines.
-    def addToLineCount (line_change)
+    def addToLineCount(line_change)
       @lines_left += line_change
     end
 
+    # If you want to tell the system how many lines are left.
+    def setLinesLeft(lines_left)
+      @lines_left = lines_left
+    end
+
     
     # Create a new format with the given top, bottom, and middle
     # formats. One argument will default to a top while two will give
@@ -710,6 +797,7 @@ module FormatR
         raise FormatException.new(),
           " You need to pass in at least one non-nil argument"
       end
+      @use_hash = false
       @io = $stdout
       @picture_lines = []
       @vars = []
@@ -747,6 +835,7 @@ module FormatR
     end
 
     # print out all the values we're holding for pictures
+    # useful for debugging 
     def showPictureLine ()
       @picture_lines.each do |line|
         puts "line:"
@@ -769,7 +858,6 @@ module FormatR
           var_name = nil
           var_name = vars.shift() unless (element.unchanging)
           var_name.strip! unless (element.unchanging)
-          #puts "got val #{val} var #{var_name}."
           output_line.push( [val,var_name] ) unless val == ""
         end
         output.push( output_line )
@@ -777,19 +865,120 @@ module FormatR
       output
     end
 
+    # if one format sets the @use_hash value everyone else will need
+    # to know too
+    def useHash (value)
+      @use_hash = value
+      @bottom.useHash( value ) unless (@bottom.nil?) 
+      @top.useHash( value ) unless (@top.nil?) 
+    end 
+
+    protected
+
+    # Print out a format using the values in the given binding output
+    # to the given io. This method will only eval the arguments on the
+    # way in and out of printing the format instead of every time they
+    # are required. This may save time but will not work for extremely
+    # dynamic formats. 
+    def printFormatFromBinding (aBinding, io = @io)
+      # save the bindings in a hash
+      collectVarValues(aBinding)
+      useHash( true )
+      @binding = @bindingVars
+      printFormatWithHash(@binding, io)
+      setVarValues(aBinding)
+    end
+
+
     # Things you shouldn't have to deal with.
     private
 
-    # When you don't want anymore on this page just fill it with blank
-    # lines and print the bottom if it's there
-    def finishPage (aBinding, suppressFF, io)
-      @binding = aBinding
-      tryOutputFormat(io)
+
+    # place the necessary variables and values into a hash so that 
+    # we can get them out at needed without having to eval every time.
+    # An interesting thing that eval gives you is JIT values. If you
+    # don't ask for a value that isn't defined you're all right.
+    def collectVarValues (aBinding) 
+      @bindingVars = Hash.new 
+      vars = getVarNames()
+      vars = vars.flatten.uniq
+
+      vars.each do |var_name| 
+        begin
+          @bindingVars[var_name] = eval( var_name, aBinding )
+        rescue NameError
+          #empty, don't bind if there is nothting there
+        end
+      end
+    end
+
+    # at the end of printing put the values back into the environment binding
+    def setVarValues (aBinding)
+      @bindingVars.each_key do |key|
+        to_eval = "#{key} = "
+        if @bindingVars[key].class == String
+          to_eval += "'#{@bindingVars[key]}'"
+        else
+          to_eval += "#{@bindingVars[key]}"
+        end
+        eval(to_eval, aBinding)
+      end
+    end
+
+    #
+    def getVarNames 
+      vars = @vars.flatten
+      vars += @top.getTopVarNames(vars) if @top
+      vars += @bottom.getBottomVarNames(vars) if @bottom
+      ##if @bottom
+      ##  res = @bottom.getBottomVarNames(vars) 
+      ##  vars += res
+      ##end
+      vars
+    end
+
+    public
+    def getTopVarNames (vars)
+      vars += @vars.flatten
+      if (@top)
+        vars += @top.getTopVarNames(vars) 
+      end
+      vars
+    end
+    
+    #
+    def getBottomVarNames (vars)
+      vars += @vars.flatten
+      if (@bottom)
+        vars += @bottom.getBottomVarNames(vars)
+      end
+      vars
+    end
+
+    private
+
+    #how large is the bottom?
+    def getBottomSize ()
       bottom_size = 0
-      (bottom_size = @bottom.getSize()) if (@bottom)
-      (@lines_left - bottom_size).times { io.puts("")}
-      @print_bottom = true
-      tryPrintBottom(io, suppressFF)
+      (bottom_size = @bottom.getSize()) if (@bottom)      
+      return bottom_size
+    end
+
+    # When you don't want anymore on this page just fill it with blank
+    # lines and print the bottom if it's there. If you've just finished
+    # a page don't bother, just do a FF if needed
+    def finishPage (aBinding, suppressFF, io)
+      if (!@print_top)
+        if (@use_hash)
+          collectVarValues(aBinding)
+          @binding = @bindingVars
+        end
+        tryOutputFormat(io)
+        bottom_size = getBottomSize();
+        (@lines_left - bottom_size).times { io.puts("")}
+        @print_bottom = true
+        tryPrintBottom(io, suppressFF)
+      end
     end
         
     # pull out the formatting
@@ -821,6 +1010,7 @@ module FormatR
           "Be sure to separate using commas:" +
           "Expected #{expected_vars} but received '#{line}'" 
       end
+      vars.collect! {|v| v.strip}
       @vars.push(vars)
       expected_vars = 0
       return expected_vars
@@ -860,7 +1050,6 @@ module FormatR
       ].each {|re|
         #num_re += 1
         if (line_section =~ re)
-          #puts "line #{line_section} matches re #{num_re}"
           output.unshift(first_char + $1)
           last_found = (first_index - 1)
           return last_found
@@ -899,7 +1088,6 @@ module FormatR
 
     # pull out from a picture line the components of the line that aren't formats
     def getNonFormats (picture_line) 
-      #puts "splitting #{picture_line}"
       lines = splitByAtOr( picture_line)
       output = []
       lines = lines.each {|element|
@@ -927,39 +1115,44 @@ module FormatR
         @buffered_lines.push("#{line}\n")
         tryPrintPartialPage(io) 
         tryPrintBottom(io)
-        tryPrintTop(io) 
-        # There is a bug with the top. When we buffer lines if the bottom
-        # won't fit we first buffer them and then print the top and bottom.
-        # the buffered lines have values before the top and bottom are printed but
-        # appear after them, ugh, it's a big mess. I can see why perl doesn't bother
-        # with bottoms!
       end
     end
 
     # if the page is too short to hold even one format just print what
-    # we can
+    # we can. True is we printed a partial page
     def tryPrintPartialPage (io)
       if (!@printed_a_body)
         bottom_size = 0
-        bottom_size = @bottom.getSize() if (@bottom)
-        if ((@buffered_lines.size() + bottom_size) == @lines_left)
-          printBufferedLines(io)
-          @print_bottom = true
-        end
+        #bottom_size = @bottom.getSize() if (@bottom)
+        #if ((@buffered_lines.size() + bottom_size) == @lines_left)
+        printBufferedLines(io)
+        @print_bottom = true
+        return true
+        #end
       end
+      return false
     end
 
     # When we have a whole format try to print it, if there isn't
     # enough room we have to save it for later.
-    def tryOutputFormat (io) 
-      bottom_size = 0
-      bottom_size = @bottom.getSize() if (@bottom)
-      #puts "buffered size #{@buffered_lines.size} bottom #{bottom_size} left: #{@lines_left}"
+    def tryOutputFormat (io, cachedBinding = nil) 
+      bottom_size = getBottomSize()
+
       if ((@buffered_lines.size() + bottom_size) <= @lines_left)
         printBufferedLines(io)
       else
-        @print_bottom = true
+        if (tryPrintPartialPage(io))
+          return true
+        else
+          unless (cachedBinding.nil?)
+            @binding = cachedBinding 
+            @buffered_lines = []
+          end
+          @print_bottom = true
+          return false
+        end
       end 
+      return true
     end
     
     #print the buffered lines
@@ -973,21 +1166,24 @@ module FormatR
     # see if a top is the right thing to print
     def tryPrintTop (io)
       if (@print_top)
+        io.print "\f" unless (@page_number == 1)
         @printed_a_body = false
-      end
-      if (@print_top && @top)
         @print_top = false
-        @top.resetPage()
-        lines = @top.printFormat(@binding, io)
-        @lines_left = (@page_length - lines)
+        if (@top)
+          @top.resetPage()
+          #lines = @top.printFormatWithHash(@binding, io)
+          lines = @top.printFormat(@binding, io)
+          @lines_left = (@page_length - lines)
+        end
       end
     end
 
     #we have a bottom, even if it's only ^L, try to get this working!
     def tryPrintBottom (io, suppressLF = false)
-      bottom_size = 0
-      bottom_size = @bottom.getSize() if (@bottom)
-
+      bottom_size = getBottomSize()
+      if ((0 == @buffered_lines.size)  && (@lines_left == bottom_size))
+        @print_bottom = true
+      end
       #this bottom is a mess if we have repeating lines
       if (@print_bottom)
         @print_bottom = false
@@ -996,7 +1192,6 @@ module FormatR
           @bottom.resetPage() 
         end
         @lines_left = @page_length
-        io.print "\f" unless suppressLF
         @page_number += 1
         @print_top = true
         return true
@@ -1004,46 +1199,98 @@ module FormatR
       return false
     end
 
-
-    # The workhorse of printFormat, buffer the format contained and
-    # then try to print it out
+    # The workhorse of the format. Print the top and bottom along with 
+    # the body. If we can't fit even one body in print as much as we can. 
+    # and split it with a top and bottom. If we can fit in one just
+    # buffer the bottom and start again on the next page.
     def printBodyFormat (io)
-      @picture_lines.each_index {|i| # There isn't a 1-1 correspondence here anymore!
+      tryPrintTop(io)
+      cachedBinding = @binding.clone
+      @picture_lines.each_index do |i| 
         line = @picture_lines[i]
         suppress = line.suppress()
         repeat = true
         while (repeat) 
-          outputLine = ""
+          if (tryPartialFullPage(io))
+            cachedBinding = @binding.clone
+          end
           vars = @vars[i].dup if @vars[i]
-          line.each {|item|
-            if (item.isUnchanging?()) 
-              outputLine << "#{item.val}"
-            else
-              begin
-                to_eval = vars.shift
-                raise FormatException.new(), 
-                  "Not enough variables supplied to match format" if (to_eval.nil?)
-                s = eval(to_eval, @binding)
-              rescue NameError
-                raise NameError.new(), "cannot find variable '#{to_eval}'"
-              end
-              suppress = false if (("" != s) && suppress) 
-              res = item.formatString(s.to_s, to_eval, @binding)
-              outputLine << "#{res}"
-            end
-          }
-          printLine(outputLine, io, suppress)
-          if ((false == suppress) && line.repeat)
+          outputLine,suppress = composeLine(line, vars, suppress)
+
+          if (!suppress)
+            outputLine.gsub!(/\s*$/, "")
+            @buffered_lines.push("#{outputLine}\n")
+          end
+
+          if ((!suppress) && line.repeat)
             suppress = line.suppress()
           else
             repeat = false
           end
+
         end #while
-      }
-      tryOutputFormat(io)
+      end # each_index
+      printed = tryOutputFormat(io,cachedBinding)
+      if (!printed)
+        @binding = cachedBinding
+        tryPrintBottom(io)
+        printBodyFormat(io)
+      end
+      tryPrintBottom(io)
       return (@page_length - @lines_left)
     end
 
+    # If we have a format that is too big to fit on a page, even the
+    # first time we will print it out along with a top and bottom and
+    # continue on. Returns if it printed or not.
+    def tryPartialFullPage (io)
+      # If we can't even fit one print all we can. 
+      if (!@printed_a_body && (@buffered_lines.size() + getBottomSize() == @lines_left))
+        tryOutputFormat(io)
+        tryPrintBottom(io)
+        tryPrintTop(io)
+        return true
+      end
+      false
+    end
+
+
+    #will shift off a var if necessary returns both a line and a value of suppress
+    def composeLine (line, vars, suppress)
+      outputLine = ""
+      line.each  do |item|
+        if (item.isUnchanging?()) 
+          outputLine << "#{item.val}"
+        else #need to chop
+          begin
+            to_eval = vars.shift
+            raise FormatException.new(), 
+              "Not enough variables supplied to match format" if (to_eval.nil?)
+            s = getVarValue( to_eval )
+          rescue NameError
+            raise NameError.new(), "cannot find variable '#{to_eval}' #{$!}"
+          end
+          suppress = false if (("" != s) && suppress) 
+          res = item.formatString(s.to_s, to_eval, @binding)
+          outputLine << "#{res}"
+        end
+      end
+      return [outputLine, suppress]
+    end
+
+
+    # Move the calls to eval into one place so we can use the hash if
+    # it's there and the binding if it's not
+    def getVarValue (var)
+      res = ""
+      if (@use_hash)
+        res = @binding[var]
+      else
+        res = eval( var, @binding )
+      end
+      return res
+    end
+
   end # class Format
 
 
@@ -1121,7 +1368,6 @@ module FormatR
         if (curval != "" && to_save =~ /^[^\s]*/)
           to_save =  " " + to_save 
         end
-        #puts "setting #{name} to #{curval + to_save}"
         @var_values[name] = curval + to_save
       else
         @var_values[name] = to_save
@@ -1141,14 +1387,13 @@ module FormatR
         matches = matches[1,matches.length]
         matches.each_with_index do |data, index|
           str, var = vars[index]
-          #puts "match str: #{str}   #{var} = #{data}"
           saveVar( str, var, data )
         end
       else
-        #no match data
         return false
+        #no match data
       end
-      true
+      return true
     end
 
     # place the values contained in the output specified by the
